'
' -----[ Title ]------------------------------------------------------
'
' File......: SNAP-024.BAS
' Purpose...: Simple 16 x 1 LCD info node for PLM-24
' Author....: Christer Johansson
' Version...: 1.02
' MCU.......: 89C2051
' Started...: 980529
' Updated...: 990130
'
' -----[ Program Description ]----------------------------------------
'
' This program shows how to implement the S.N.A.P protocol in BASCOM
' and is an simple example to receive data from another node and
' display it on the LCD. It uses PLM-24 Power Line Modems to 
' communicate over the power lines (mains). This example uses 16-bit
' CRC-CCITT as error detection method which gives secure data transfer.
'
' If the node is addressed by another node (PC or another MCU) it shows
' what node sent the packet and the value of DB1 in hexadecimal on the
' LCD. In the example below node with address &H05 is sending &HEF in
' the data byte.
'
'   16x1 LCD display
' +------------------+
' | FROM:05  DATA:EF |
' +------------------+
'
' The packet structure is defined in the received packets first two
' bytes (HDB2 and HDB1). The following packet structure is used.
'
' DD=01     - 1 Byte destination address
' SS=01     - 1 Byte source address
' PP=00     - No protocol specific flags
' AA=01     - Acknowledge is required
' C=0       - Command mode not supported
' EEE=100   - 16-bit CRC-CCITT
' NNNN=0001 - 1 Byte data
'
' Overview of header definition bytes (HDB2 and HDB1)
'
'        HDB2              HDB1
' +-----------------+-----------------+
' | D D S S P P A A | C E E E N N N N |
' +-----------------+-----------------+
'
' For more information about PLM-24 Power Line Modem see URL below.
'
' http://www.hth.com/plm-24/
'
' For more information about S.N.A.P see URL below.
'
' http://www.hth.com/snap/
'
' Below is an short overview of the S.N.A.P network protocol.
'
' Features
' --------
'
' - Free and open network protocol.
' - Easy to learn, use and implement.
' - Up to 16.7 million destination node addresses.
' - Up to 16.7 million source node addresses.
' - 0-24 protocol specific flags.
' - Optional ACK/NAK request.
' - Optional command mode.
' - 8 different error detecting methods.
' - Header is scaleable from 3-14 bytes.
' - 0-512 Bytes data in each packet.
' - Can be used as master/slave or peer-to-peer.
' - Supports broadcast messages.
'
' -----[ Disclaimer ]-------------------------------------------------
' 
' This example is offered on an "AS IS" basis, no warranty expressed
' or implied. HTH disclaim liability of any damages associated with
' the use of the hardware or software described herein. You use it on
' your own risk. We are not able to provide any free support.
'
' Copyright (c) 1998-1999 HTH all rights reserved
'
' -----[ Revision History ]-------------------------------------------
'
' 980529 - Version 1.00    Initial release
' 980919 - Version 1.01    Revised for S.N.A.P Version 0.91
' 990130 - Version 1.02    Revised for S.N.A.P Version 1.00
'
' -----[ Aliases ]----------------------------------------------------
'
CD alias P3.2								' PLM-24 carrier detect pin
RxTx alias P3.7                     ' PLM-24 Rx/Tx control pin
SPKR alias P1.0							' Speaker output pin

' -----[ Constants ]--------------------------------------------------
'
dim Preamble as const &B01010101		' Preamble byte
dim SYNC as const &B01010100			' Synchronisation byte
dim CRCPOLY as const &H1021			' CRC-CCITT
dim MyAddress as const &H04			' Address for this node (1-255)

' -----[ Variables ]--------------------------------------------------
'
dim TEMP1 as byte							' Temporary Variable
dim TEMP2 as byte							' Temporary Variable
dim TEMP_CRC as word						' Temporary Variable		
dim CRC as word							' CRC word variable
dim CRC2 as byte							' CRC hi byte variable
dim CRC1 as byte							' CRC lo byte variable
dim HDB2 as byte							' Header Definition Byte 2
dim HDB1 as byte							' Header Definition Byte 1
dim DB1 as byte							' Packet data
dim SAB1 as byte							' What node sent this packet
dim DAB1 as byte							' What node should have this paket

' -----[ Initialization ]---------------------------------------------
'
' Define serial communication parameters
$baud = 2400
$crystal = 11059200

' Configure LCD display
config lcd = 16 * 1

cls											' Clear the LCD display
cursor off noblink						' Hide cursor
set RxTx										' Set PLM-24 to receive mode
set CD										' Make CD pin as input

'------[ Program ]----------------------------------------------------
'

_Start:

	' Wait for data on serialport
	TEMP1 = waitkey

	' If received data is a SYNC byte read next seven bytes
	' from master, if not return to start
	if TEMP1 <> SYNC then
		goto _Start
	else
		' Get packet in binary mode
		inputbin HDB2,HDB1,DAB1,SAB1,DB1,CRC2,CRC1

		'
		' Packet header check routine
   	'
		' Check HDB2 to see if MCU are capable to use the packet
		' structure, if not goto Start
		if HDB2 <> &B01010001 then
			goto _Start
		end if
		' Check HDB1 to see if MCu are capable to use the packet
   	' structure, if not goto Start
		if HDB1 <> &B01000001 then
			goto _Start
		end if

		'
		' Address check routine
   	'
   	' Check if this is the node addressed, if not goto Start
   	if DAB1 <> MyAddress then
			goto _Start
		end if

		'
		' CRC check routine
		'
		' Calculate CRC for all received bytes
		CRC = 0
		Temp2 = HDB2
		gosub calc_CRC
   	Temp2 = HDB1
		gosub calc_CRC
   	Temp2 = DAB1
   	gosub calc_CRC
   	Temp2 = SAB1
   	gosub calc_CRC
   	Temp2 = DB1
   	gosub calc_CRC
   	Temp2 = CRC2 
   	gosub calc_CRC
   	Temp2 = CRC1
   	gosub calc_CRC
   	' Check if there was any CRC errors, if so send NAK
		if CRC <> 0 then
			goto _NAK
		end if

   	' No CRC errors in packet so clear the LCD and show
		' data and from what node it was sent
		cls
		lcd "FROM:"
		lcdhex SAB1
		lcd "  "
		lcd "DATA:"
		lcdhex DB1
		' Beep to alert new message
		sound SPKR , 10000, 10
		' Clear DB1
		DB1 = 0
		' Send ACK
		goto _ACK

_ACK:
        	' Send ACK (i.e tell master that packet was OK)
        	' Set ACK bits to ACK in HDB2 (logic XOR bit0 and bit1)
		HDB2 = HDB2 xor &B00000011
		goto _Send

_NAK:
        	' Send NAK (i.e tell master that packet was bad)
        	' Set ACK bits to NAK in HDB2 (logic OR bit0 and bit1)
		HDB2 = HDB2 or &B00000011
		goto _Send

_Send:
		' Swap SAB1 <-> DAB1 address bytes
		swap SAB1, DAB1   

		' Clear CRC variable
		CRC = 0
		' Put HDB2 in tempory variable
		Temp2 = HDB2
		' Calculate CRC
		gosub calc_crc
        
		' Put HDB1 in tempory variable
		Temp2 = HDB1
		' Calculate CRC
		gosub calc_crc
        
		' Put DAB1 in tempory variable
		Temp2 = DAB1
		' Calculate CRC
		gosub calc_crc

		' Put SAB1 in tempory variable
		Temp2 = SAB1
		' Calculate CRC
		gosub calc_crc

		' Put DB1 in tempory variable
		Temp2 = DB1
		' Calculate CRC
		gosub calc_crc

		' Put calculated Hi_CRC value in CRC2
		CRC2 = high(CRC)
		' Put calculated Lo_CRC value in CRC1
		CRC1 = low(CRC)
        
_Check_CD:
		' Check carrier detect pin if some other PLM-24 unit is sending       
		if CD = 1 then
			goto _Tx
		else
			' Someone else was sendning. Wait a "semi" random time (0-255 ms) 
			' before trying to send again
			waitms CRC1
			goto _Check_CD
		end if

_Tx:
		' Set PLM-24 Power Line Modem in transmit mode
		reset RxTx
		' Give PLM-24 time to settle before transmitt any data
		waitms 5

		' Send packet to master, including the SYNC byte
		print chr(Preamble);chr(SYNC);chr(HDB2);chr(HDB1);chr(DAB1);chr(SAB1);chr(DB1);chr(CRC2);chr(CRC1);

		' Give STAMP time to shift out all bits before setting to Rx
   	waitms 50
		' Set PLM-24 Power Line Modem in receive mode
		set RxTx

		' Done, go back to Start and wait for a new packet
		goto _Start

		end if

' -----[ Subroutines ]------------------------------------------------
'
' Subroutine for calculating CRC value in variable Temp2 
calc_crc:
	TEMP_CRC = Temp2 * 256
	CRC = TEMP_CRC xor CRC
	for Temp1 = 0 to 7
		if CRC > 32767 then
			CRC = CRC * 2
			CRC = CRC xor CRCPOLY
		else
			CRC = CRC * 2
		end If
			CRC = CRC and 65535
	next
return

' -----[ End ]--------------------------------------------------------

